@@ -21,9 +21,18 @@
 #include <sys/ioctl.h>
 #include <linux/tiocl.h>
 #include <stdint.h>
+#include <linux/kd.h>
 
 #include "consolation.h"
 
+static int
+check_mode(int fd)
+{
+  int mode;
+  ioctl(fd, KDGETMODE, &mode);
+  return mode==KD_TEXT;
+}
+
 void
 set_screen_size(void)
 {
@@ -47,12 +56,13 @@ linux_selection(int xs, int ys, int xe, int ye, int sel_mode)
   s.argp[1] = TIOCL_SETSEL;
   s.sel.xs = xs;
   s.sel.ys = ys;
-  s.sel.xe = xe;
-  s.sel.ye = ye;
+  s.sel.xe = xe<0 ? xs: xe;
+  s.sel.ye = ye<0 ? ys: ye;
   s.sel.sel_mode = sel_mode;
   fd = open("/dev/tty0",O_RDONLY);
-  if(ioctl(fd, TIOCLINUX, ((char*)&s)+1)<0)
-    perror("selection: TIOCLINUX");
+  if (check_mode(fd))
+    if (ioctl(fd, TIOCLINUX, ((char*)&s)+1)<0)
+      perror("selection: TIOCLINUX");
   close(fd);
 }
 
@@ -65,30 +75,28 @@ draw_pointer(int x, int y)
 void
 select_region(int x, int y, int x2, int y2)
 {
-  if (x2 < 0) x2 = x;
-  if (y2 < 0) y2 = y;
   linux_selection(x, y, x2, y2, TIOCL_SELCHAR);
 }
 
 void
-select_word(int x, int y)
+select_words(int x, int y, int x2, int y2)
 {
-  linux_selection(x, y, x, y, TIOCL_SELWORD);
+  linux_selection(x, y, x2, y2, TIOCL_SELWORD);
 }
 
 void
-select_line(int x, int y)
+select_lines(int x, int y, int x2, int y2)
 {
-  linux_selection(x, y, x, y, TIOCL_SELLINE);
+  linux_selection(x, y, x2, y2, TIOCL_SELLINE);
 }
 
 void paste(void)
 {
-  int fd;
   char subcode = TIOCL_PASTESEL;
-  fd = open("/dev/tty0",O_RDWR);
-  if(ioctl(fd, TIOCLINUX, &subcode)<0)
-    perror("paste: TIOCLINUX");
+  int fd = open("/dev/tty0",O_RDWR);
+  if (check_mode(fd))
+    if (ioctl(fd, TIOCLINUX, &subcode)<0)
+      perror("paste: TIOCLINUX");
   close(fd);
 }
 
@@ -105,8 +113,9 @@ void scroll(int sc)
   scr.subcode[3] = 0;
   scr.sc = sc;
   fd = open("/dev/tty0",O_RDONLY);
-  if(ioctl(fd, TIOCLINUX, &scr)<0)
-    perror("scroll: TIOCLINUX");
+  if (check_mode(fd))
+    if (ioctl(fd, TIOCLINUX, &scr)<0)
+      perror("scroll: TIOCLINUX");
   close(fd);
 }
 
@@ -153,7 +162,8 @@ void set_lut(const char *def)
     }
   }
   fd = open("/dev/tty0",O_RDWR);
-  if(ioctl(fd, TIOCLINUX, &l)<0)
-    perror("set_lut: TIOCLINUX");
+  if (check_mode(fd))
+    if (ioctl(fd, TIOCLINUX, &l)<0)
+      perror("set_lut: TIOCLINUX");
   close(fd);
 }
